#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <machine/param.h>

#include <machine/db_machdep.h>
#define ELFSIZE DB_ELFSIZE

#include <sys/exec_elf.h>
#include <machine/bootinfo.h>
#include "memio.h"

int
scanfile(fd, start, end, entry, load)
	int fd, load;
	caddr_t *start, *end, *entry;
{
	Elf_Ehdr elfx, *elf = &elfx;
	int i, first;
	Elf_Shdr *shtbl = NULL;
	Elf_Phdr *phtbl = NULL;
	caddr_t min_addr, max_addr;
	char page_buf[NBPG];

	if (lseek(fd, 0, SEEK_SET) == -1)  {
		fprintf(stderr,"seek error\n");
		goto error_cleanup;
	}
	if (read(fd, (void*)elf, sizeof(Elf_Ehdr)) != sizeof(Elf_Ehdr)) {
		fprintf(stderr,"read header error\n");
		goto error_cleanup;
	}

	if ((phtbl = (Elf_Phdr *)malloc(sizeof(*phtbl) * elf->e_phnum)) == NULL
	|| (shtbl = (Elf_Shdr *)malloc(sizeof(*shtbl) * elf->e_shnum)) == NULL) {
		fprintf(stderr,"malloc() error\n");
		goto error_cleanup;
	}

	if (lseek(fd, elf->e_phoff, SEEK_SET) == -1)  {
		fprintf(stderr,"seek for program header table error\n");
		goto error_cleanup;
	}
	if (read(fd, (void *)phtbl, sizeof(Elf_Phdr) * elf->e_phnum)
			!= (int)(sizeof(Elf_Phdr) * elf->e_phnum)) {
		fprintf(stderr,"read program header table error\n");
		goto error_cleanup;
	}

	if (lseek(fd, elf->e_shoff, SEEK_SET) == -1)  {
		fprintf(stderr,"seek for segment header table error.\n");
		goto error_cleanup;
	}
	if (read(fd, (void *)shtbl, sizeof(Elf_Shdr) * elf->e_shnum)
			!= (int)(sizeof(Elf_Shdr) * elf->e_shnum)) {
		fprintf(stderr,"read segment header table error\n");
		goto error_cleanup;
	}

	first = 1;
	for (i = 0; i < elf->e_phnum; i++) {
		if (phtbl[i].p_type != PT_LOAD) {
			continue;
		}

		if (first || max_addr
			< (caddr_t)(phtbl[i].p_vaddr + phtbl[i].p_memsz)) {
			max_addr
			= (caddr_t)(phtbl[i].p_vaddr + phtbl[i].p_memsz);
		}
		if (first || (caddr_t)phtbl[i].p_vaddr < min_addr) {
			min_addr = (caddr_t)phtbl[i].p_vaddr;
		}

		if (load) {
			vaddr_t file_end, mem_end, offset;
			int load_len, fill_len;
			size_t len;

			if (lseek(fd, phtbl[i].p_offset, SEEK_SET) == -1)  {
				goto error_cleanup;
			}

			offset = phtbl[i].p_vaddr;
			if (lseek(load, offset, SEEK_SET) == -1)  {
				goto error_cleanup;
			}
			file_end = offset + phtbl[i].p_filesz;
			mem_end = offset + phtbl[i].p_memsz;
			while (offset < mem_end) {
				vaddr_t next_offset;

				next_offset = offset + NBPG < mem_end ? 
				    offset + NBPG : mem_end;
				len = next_offset - offset;
				load_len = file_end < offset ? 0 : file_end - offset;
				if (load_len < NBPG)
					fill_len = len - load_len;
				else {
					load_len = NBPG;
					fill_len = 0;
				}
				if (read(fd, page_buf, load_len) != load_len)  {
					fprintf(stderr, "can't load kernel to buf\n");
					goto error_cleanup;
				}
				memset(page_buf + load_len, 0, fill_len);
				if (write(load, page_buf, len) != len) {
					fprintf(stderr, "can't write kernel to akmem\n");
					goto error_cleanup;
				}
				offset = next_offset;
			}
		}
		first = 0;
	}
	if (first) {
		goto error_cleanup;
	}

	if (phtbl) free(phtbl);
	if (shtbl) free(shtbl);

	if (start) *start = min_addr;
	if (end) *end = max_addr;
	if (entry) *entry = (caddr_t)elf->e_entry;
	return (0);

error_cleanup:
	if (phtbl) free(phtbl);
        if (shtbl) free(shtbl);
	return (1);
}

int
main(int argc, char **argv)
{
	int kernel, akmem;
	struct mem_ak_alloc alloc_info;
	struct mem_ak_exec exec_info;
	caddr_t start, end, entry;

	if (argc < 2) {
		fprintf(stderr, "please new kernel\n");
		exit(1);
	}
	if ((kernel = open(argv[1], O_RDONLY)) < 0) {
		fprintf(stderr, "can't open %s\n", argv[1]);
		exit(1);
	}
	fprintf(stderr, "rebooting %s...\n", argv[1]);
	scanfile(kernel, &start, &end, &entry, 0);
	exec_info.entry = entry;

	if ((akmem = open("/dev/akmem", O_RDWR)) < 0) {
		fprintf(stderr, "can't open /dev/akmem\n");
		exit(1);
	}

	alloc_info.nsegments = 1;
	alloc_info.segments = (struct akmem_segment*)
	    malloc(alloc_info.nsegments * sizeof(struct akmem_segment));
	alloc_info.segments[0].start = (vaddr_t)start;
	alloc_info.segments[0].end = (vaddr_t)end;

	if (ioctl(akmem, MEMIOCTL_AKALLOC, &alloc_info) < 0) {
		fprintf(stderr, "can't allocate physical memory\n");
		close(akmem);
		close(kernel);
		exit(1);
	}
	if (scanfile(kernel, &start, &end, &entry, akmem) != 0) {
		fprintf(stderr, "can't load kernel\n");
		close(kernel);
		close(akmem);
		exit(1);
	}
	close(kernel);
	exec_info.argc = argc - 1;
	exec_info.argv = argv + 1;
	exec_info.env = NULL;
	ioctl(akmem, MEMIOCTL_AKEXEC, &exec_info);
	/* not reachable */
	exit(1);
}
